// NOTICE: 

//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


///////////////////////////////////////////////////////////////////////////
// HEADERS
///////////////////////////////////////////////////////////////////////////

#define IDC_POSITION		(500)
//#define LVM_GETSELECTIONMARK	(LVM_FIRST+66)


#include <windows.h>
#include <stdio.h>
#include "resource.h"
#include "..\include\Filter.h"
#include <commctrl.h>
#include <string.h>
//#include <Shlwapi.h>

//#include "..\include\StdAfx.h"
#include "..\include\ScriptInterpreter.h"
#include "..\include\ScriptError.h"
#include "..\include\ScriptValue.h"


///////////////////////////////////////////////////////////////////////////
// FUNCTION DEFS
///////////////////////////////////////////////////////////////////////////

int panaRunProc(const FilterActivation *fa, const FilterFunctions *ff);
int panaStartProc(FilterActivation *fa, const FilterFunctions *ff);
int panaEndProc(FilterActivation *fa, const FilterFunctions *ff);
long panaParamProc(FilterActivation *fa, const FilterFunctions *ff);
int panaConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void panaStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
void panaScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool panaFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);
void xRectFill(const FilterActivation *fa, PixCoord x, PixCoord y, PixDim dx, PixDim dy, Pixel32 c);
void rebuildListView(HWND hListbox, void *filterData);
void deleteFaderStep(void *filterData, int index);

LRESULT CALLBACK nPreviewWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);


///////////////////////////////////////////////////////////////////////////
// DATA DEFS
///////////////////////////////////////////////////////////////////////////

//****************************************
// Structure where we hold our Global 
// filter specific Data, like dialog settings,
// and lookup tables.
//
// This is passed into every function ....
//****************************************
typedef struct panStep {
	int frameNumber;
	POINT ptUL;
} panStep;

typedef struct MyFilterData {

	POINT ptMove;

	struct panStep **stepData;
	int stepCount;
	int foundItem;
	panStep prevStep;
	panStep nextStep;


	int dstW;
	int dstH;
	long frame;
	long saveFrame;
	int mode;
	int inverse;
	int drawing;

	HWND fpHwnd;
	HWND dlgHwnd;
	WNDPROC fpWndProc;
	HCURSOR hCsr;

	IFilterPreview *ifp;
	FilterStateInfo *fsi;

} MyFilterData;

MyFilterData *g_mfd;




//****************************************
// ?????
//****************************************
ScriptFunctionDef pana_func_defs[]={
    { (ScriptFunctionPtr)panaScriptConfig, "Config", "0iiii" },
    { NULL },
};


//****************************************
// ?????
//****************************************
CScriptObject pana_obj={
    NULL, pana_func_defs
};

//****************************************
// Structure that tells vDub about our filter
// Holds name, description, and pointers to
// the functions vdub calls for the filter
// to do it's thing.
//****************************************
struct FilterDefinition filterDef_pana = {

    NULL, NULL, NULL,       // next, prev, module
    "pan N scan - alpha",             // name
    "Reposition frame between key frames.",
                            // desc
    "Loadus",               // maker
    NULL,                   // private_data
    sizeof(MyFilterData),   // inst_data_size

    NULL,                   // initProc
    NULL,                   // deinitProc
    panaRunProc,        // runProc
    panaParamProc,      // paramProc
    panaConfigProc,     // configProc
    panaStringProc,     // stringProc
    panaStartProc,      // startProc
    panaEndProc,        // endProc

    &pana_obj,          // script_obj
    panaFssProc,        // fssProc

};

static FilterDefinition *fd_pana;


///////////////////////////////////////////////////////////////////////////
// FUNCTIONS
///////////////////////////////////////////////////////////////////////////


//****************************************
// VDub calls this exported function to 
// Add our filter to the list ..
//****************************************

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
    if (!(fd_pana = ff->addFilter(fm, &filterDef_pana, sizeof(FilterDefinition))))
        return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

    return 0;
}

//****************************************
// Exported function to remove us from the list
//****************************************
void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
    ff->removeFilter(fd_pana);
}


//****************************************
// StartProc - Just before we start to filter
//
// Good place to .....
//****************************************
int panaStartProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    return 0;
}

int setUpPan(const FilterActivation *fa) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	int  i;



	//*** Catch invalid cases ***
	if(mfd->stepCount < 1	//-- Less than 1 step
		|| mfd->saveFrame > mfd->stepData[mfd->stepCount - 1]->frameNumber) {  //-- After last step

		return 0;
	}

	//--- Already set up ---
	if(mfd->saveFrame > mfd->prevStep.frameNumber 
		&& mfd->saveFrame <= mfd->nextStep.frameNumber
		&& mfd->saveFrame != 0) {
		return 1;
	}


	//**** Find PrevStep/NextStep ****

	for( i=0;i<mfd->stepCount-1;i++) {
		if(mfd->stepData[i]->frameNumber >= mfd->saveFrame) {
			break;
		}
	}

	if(i > 0) {

		//*** Found a step before & after current frame
		mfd->prevStep.frameNumber = mfd->stepData[i-1]->frameNumber;
		mfd->prevStep.ptUL.x = mfd->stepData[i-1]->ptUL.x;
		mfd->prevStep.ptUL.y = mfd->stepData[i-1]->ptUL.y;
	
		mfd->nextStep.frameNumber = mfd->stepData[i]->frameNumber;
		mfd->nextStep.ptUL.x = mfd->stepData[i]->ptUL.x;
		mfd->nextStep.ptUL.y = mfd->stepData[i]->ptUL.y;
	} else {
		//*** Found a step after but not before so fake it
		mfd->prevStep.frameNumber = 0;
		mfd->prevStep.ptUL.x = 0;
		mfd->prevStep.ptUL.y = 0;
	
		mfd->nextStep.frameNumber = mfd->stepData[i]->frameNumber;
		mfd->nextStep.ptUL.x = mfd->stepData[i]->ptUL.x;
		mfd->nextStep.ptUL.y = mfd->stepData[i]->ptUL.y;
	}

	return 1;

}

//****************************************
// RunProc gets called for each frame.
//
// This is where the filtering is done.
//****************************************

int panaRunProc(const FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;
    //PixDim w, h;
    Pixel32 *src, *dst;
	double ratio;
	POINT cur;
	cur.x = 0; cur.y=0;


	mfd->saveFrame = fa->pfsi->lCurrentFrame;

    src = (Pixel32 *)fa->src.data;
    dst = (Pixel32 *)fa->dst.data;


	if(setUpPan(fa)) {

		ratio = (double)(mfd->saveFrame - mfd->prevStep.frameNumber) 
			/ (double)(mfd->nextStep.frameNumber - mfd->prevStep.frameNumber); 
		cur.x = int(mfd->prevStep.ptUL.x 
			+ (mfd->nextStep.ptUL.x - mfd->prevStep.ptUL.x) * ratio);
		cur.y = int(mfd->prevStep.ptUL.y 
			+ (mfd->nextStep.ptUL.y - mfd->prevStep.ptUL.y) * ratio);

		fa->dst.RectFill(0,0,-1,-1,0);
		fa->dst.BitBlt(0,0,&(fa->src)
				,cur.x
				,cur.y,-1,-1);
	} else {
			fa->dst.BitBlt(0,0,&(fa->src)
				,0
				,0,-1,-1);

	}

	return 0;
}


//****************************************
// EndProc - Done Filtering 
//
// Good place to clean up 
//****************************************
int panaEndProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    return 0;
}


//****************************************
// ParamProc - Set the Frame size
//
// Good place to .....
//****************************************
long panaParamProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	fa->dst.pitch = (fa->dst.w*4 + 7) & -8;

	mfd->dstW = fa->dst.w;
	mfd->dstH = fa->dst.h;


    return FILTERPARAM_SWAP_BUFFERS;
}


//****************************************
// ConfigDlgProc - Controls our configuration dialog
//
// This is where you need to do windows programing.
//****************************************
BOOL CALLBACK ConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
    MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);

	static	char msgDraw[] = "Left click and drag in the preview window to draw the mask.";
	static	char msgMove[] = "Left click on the mask, and drag to move it in the preview window.";

	HWND hListBox = GetDlgItem(hdlg, IDC_LIST);
	LVCOLUMN column = {LVCF_FMT | LVCF_TEXT | LVCF_WIDTH, LVCFMT_CENTER, 70, "Frame", 0, 0, -1, -1};
	int indexLV;


	//

    switch(msg) {
        case WM_INITDIALOG:
            SetWindowLong(hdlg, DWL_USER, lParam);
            mfd = (MyFilterData *)lParam;

			mfd->dlgHwnd = hdlg;

			ListView_SetExtendedListViewStyle(hListBox, ListView_GetExtendedListViewStyle(hListBox)| LVS_EX_FULLROWSELECT);

			// now create the columns
			ListView_InsertColumn(hListBox, 0, &column); // dummy column!
			ListView_InsertColumn(hListBox, 1, &column);
			column.pszText = "Top"; column.cx = 50;
			ListView_InsertColumn(hListBox, 2, &column);
			column.pszText = "Left"; 
			ListView_InsertColumn(hListBox, 3, &column);
			ListView_DeleteColumn(hListBox, 0);
			rebuildListView(hListBox, mfd);
			
			if (mfd->ifp) {
				mfd->ifp->InitButton(GetDlgItem(hdlg, IDC_PREVIEW));
				mfd->ifp->Toggle(hdlg);
			}



            return TRUE;

        case WM_COMMAND:
            switch(LOWORD(wParam)) {
            case IDOK:
                EndDialog(hdlg, 0);
                return TRUE;
            case IDCANCEL:
                EndDialog(hdlg, 1);
                return FALSE;
			case IDC_PREVIEW:
				if (mfd->ifp)
					mfd->ifp->Toggle(hdlg);
				return FALSE;

			case IDC_GOTO:
				if(mfd->ifp) {
					indexLV = SendMessage( hListBox, LVM_GETSELECTIONMARK,0,0 );
					if(indexLV != -1) {
						SendMessage(GetDlgItem(mfd->fpHwnd, IDC_POSITION), (WM_USER+0x103), (WPARAM)TRUE, mfd->stepData[indexLV]->frameNumber);				
						mfd->ifp->RedoFrame();
					}
				}
				return FALSE;



			case IDC_REMOVE: 
					indexLV = SendMessage( hListBox, LVM_GETSELECTIONMARK,0,0 );
					if(indexLV != -1) {
					deleteFaderStep(mfd, indexLV);
					rebuildListView(hListBox, mfd);
				}
				return FALSE;
            }
            break;
    }

    return FALSE;
}


void rebuildListView(HWND hListbox, void *filterData) {
    MyFilterData *mfd = (MyFilterData *)filterData;
	LVITEM item = {LVIF_PARAM | LVIF_TEXT, 0, 0, 0, 0, NULL, 0, NULL, 0, 0};
	ListView_DeleteAllItems(hListbox);

	for (int i = 0; i < mfd->stepCount; i++) {
		char t[10];
		struct panStep *step;

		step = mfd->stepData[i];

		sprintf(t, "%lu", step->frameNumber);
		if (item.pszText) {
			free(item.pszText);
		}
		item.mask |= LVIF_PARAM;
		item.pszText = (char *)malloc(strlen(t) + 1);
		strcpy(item.pszText, t);
		item.lParam = step->frameNumber;
		item.iItem = i;
		item.iSubItem = 0;
		item.iItem = ListView_InsertItem(hListbox, &item);
		
		sprintf(t, "%d", step->ptUL.y);
		item.mask &= ~LVIF_PARAM;
		if (item.pszText) {
			free(item.pszText);
		}
		item.pszText = (char *)malloc(strlen(t) + 1);
		strcpy(item.pszText, t);
		item.iSubItem = 1;
		ListView_SetItem(hListbox, &item);


		sprintf(t, "%d", step->ptUL.x);
		item.mask &= ~LVIF_PARAM;
		if (item.pszText) {
			free(item.pszText);
		}
		item.pszText = (char *)malloc(strlen(t) + 1);
		strcpy(item.pszText, t);
		item.iSubItem = 2;
		ListView_SetItem(hListbox, &item);

	}
}


BOOL insertStep(void *filterData, int frame, int x, int y) {
    MyFilterData *mfd = (MyFilterData *)filterData;

	int step = 0;
	unsigned char shift = 0;
	struct panStep **n = (panStep **)malloc((sizeof(struct panStep *) * (mfd->stepCount + 1)));
	struct panStep *nstep = (struct panStep *)malloc(sizeof(struct panStep));



	if (frame < 1)
		return FALSE;

	mfd->stepCount++;

	if (mfd->stepCount == 1) {
		nstep->frameNumber = frame;
		nstep->ptUL.x = x;
		nstep->ptUL.y = y;
		n[0] = nstep;
	} else {
		for (int i = 0; i < mfd->stepCount - 1; i++)
			if (frame == mfd->stepData[i]->frameNumber) {
				mfd->stepCount--;
				return FALSE;
			}
		while (step < mfd->stepCount) {
			if (!shift && ((step == mfd->stepCount - 1) 
				|| (mfd->stepData[step]->frameNumber > frame))) {

				nstep->frameNumber = frame;
				nstep->ptUL.x = x;
				nstep->ptUL.y = y;
				n[step] = nstep;
				shift = 1;
			} else {
				n[step] = mfd->stepData[step - shift];
			}
			step++;
		}
	}
	if (mfd->stepData)
		free(mfd->stepData);
	mfd->stepData = n;
	return TRUE;
}


void deleteFaderStep(void *filterData, int index) {
    MyFilterData *mfd = (MyFilterData *)filterData;
	
	struct panStep **n = (panStep **)malloc((sizeof(struct panStep *) * (mfd->stepCount - 1)));
	int shift = 0;
	for (int i = 0; i < mfd->stepCount - 1; i++) {
		if (index == i)
			shift = 1;
		n[i] = mfd->stepData[i + shift];
	}
	mfd->stepCount--;
	free(mfd->stepData);
	mfd->stepData = n;
}
//****************************************
// fPreviewButtonCallback - Vdub calls when preview is opened/closed
//
// Turn on/off capture of preview window mouse events
//****************************************

void *fPreviewButtonCallback(bool fNewState, void *pData) {

	MyFilterData *mfd = (MyFilterData *) pData;
	g_mfd = mfd;


	if(fNewState) {
		mfd->fpHwnd = FindWindow(0,"Filter Preview");

		//*** Install filter proc for preview windows messages ....
		mfd->fpWndProc = (WNDPROC) SetWindowLong(mfd->fpHwnd, GWL_WNDPROC, (LONG) nPreviewWndProc );

	} else {

		//*** restore proc for filter preview windows messages ....
		mfd->fpWndProc = (WNDPROC) SetWindowLong(mfd->fpHwnd, GWL_WNDPROC, (LONG) mfd->fpWndProc );
	}
	return FALSE;
}



//****************************************
// ConfigProc - Just before our dialog box is opened
//
// This is where you .....
//****************************************
int panaConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	//---- Setup preview w/ callback on toggle
	mfd->ifp = fa->ifp;
	mfd->ifp->SetButtonCallback((FilterPreviewButtonCallback)fPreviewButtonCallback, mfd); 


    return DialogBoxParam(fa->filter->module->hInstModule,
            MAKEINTRESOURCE(IDD_FILTER_PANA), hwnd,
            (DLGPROC) ConfigDlgProc, (LPARAM)fa->filter_data);
}


//****************************************
// StringProc - Called to get onfiguration settings you
// see in the filters config list.
//
// Basically builds a string of the current settings.
//****************************************
void panaStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str) {
    //char *desc[50];
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

}


//****************************************
// ScriptConfig - Sets your settings from a list
// of script arguments.
//****************************************
void panaScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
    FilterActivation *fa = (FilterActivation *)lpVoid;
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

}

//****************************************
// FssProc - Saves your args as a string for scripts.
//****************************************
bool panaFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;


    return true;
}


////////////////////////////////////////////
// Msg Proc to process filter preview windows messages
////////////////////////////////////////////

LRESULT CALLBACK nPreviewWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {

	HWND hListBox = GetDlgItem(g_mfd->dlgHwnd, IDC_LIST);

    POINTS ptLBD;
	POINT ptLBD2;
    RECT rcClient;                 // client area rectangle 
    POINT ptClientUL;              // client upper left corner 
    POINT ptClientLR;              // client lower right corner 

	int foundFlag, i;
	foundFlag = 0;

	switch(msg) {

        case WM_MOUSEMOVE: 
 
            if (wParam & MK_LBUTTON)  { 


				if(g_mfd->drawing == TRUE) {

					ptLBD = MAKEPOINTS(lParam); 
					g_mfd->stepData[g_mfd->foundItem]->ptUL.x = 
						(g_mfd->ptMove.x - ptLBD.x );
					g_mfd->stepData[g_mfd->foundItem]->ptUL.y = 
						( g_mfd->ptMove.y - ptLBD.y);

					g_mfd->nextStep.ptUL.x = g_mfd->stepData[g_mfd->foundItem]->ptUL.x;
					g_mfd->nextStep.ptUL.y = g_mfd->stepData[g_mfd->foundItem]->ptUL.y;

					g_mfd->ifp->RedoFrame();


				}
			}
			break;

		case WM_LBUTTONDBLCLK:
			ptLBD = MAKEPOINTS(lParam); 

			if(insertStep(g_mfd,  g_mfd->saveFrame, 0, 0)) {
				rebuildListView(hListBox, g_mfd);
			}


			break;

       case WM_LBUTTONDOWN: 
 
		ptLBD = MAKEPOINTS(lParam); 
		POINTSTOPOINT(ptLBD2,ptLBD);


		for( i=0;i<g_mfd->stepCount;i++) {
			if(g_mfd->stepData[i]->frameNumber == g_mfd->saveFrame) {
				foundFlag = 1;
				g_mfd->foundItem = i;
				break;
				
			}
		}
		if(foundFlag) {
				g_mfd->drawing = TRUE;
				g_mfd->ptMove.x = ptLBD.x + g_mfd->stepData[g_mfd->foundItem]->ptUL.x;
				g_mfd->ptMove.y = ptLBD.y + g_mfd->stepData[g_mfd->foundItem]->ptUL.y;
		
			//**** If click in client area start draw 
			if(ptLBD.x-3 > 0 
				&& ptLBD.x-3 < g_mfd->dstW 
				&& ptLBD.y-3 > 0 
				&& ptLBD.y-3 < g_mfd->dstH) {


				GetClientRect(hWnd, &rcClient); 
				ptClientUL.x = rcClient.left + 4; 
				ptClientUL.y = rcClient.top + 4; 
				ptClientLR.x = ptClientUL.x + g_mfd->dstW; 
				ptClientLR.y = ptClientUL.y + g_mfd->dstH; 

				ClientToScreen(hWnd, &ptClientUL); 
				ClientToScreen(hWnd, &ptClientLR); 
 
				SetRect(&rcClient, ptClientUL.x, ptClientUL.y, ptClientLR.x, ptClientLR.y); 
				ClipCursor(&rcClient); 

			}
		}
			
		break;

       case WM_LBUTTONUP: 


		   if(g_mfd->drawing) {
			ClipCursor(NULL); 
			g_mfd->drawing = FALSE;
			g_mfd->ifp->RedoFrame();
			rebuildListView(hListBox, g_mfd);

		   }

	}

	return CallWindowProc( g_mfd->fpWndProc, hWnd, msg, wParam, lParam);
//	return CallWindowProc((DLGPROC) g_mfd->fpWndProc, hWnd, msg, wParam, lParam);


}


//*********************************

void xRectFill(const FilterActivation *fa, PixCoord x, PixCoord y, PixDim dx, PixDim dy, Pixel32 c) {


	// Do the blit

	Pixel32 *dst = (Pixel32 *)fa->dst.data;
	int w = fa->dst.w;
	int h = fa->dst.h;
	int m = fa->dst.modulo;
	PixOffset pitch = fa->dst.pitch;



	if (dx == -1) dx = w;
	if (dy == -1) dy = h;

	// clip to destination bitmap

	if (x < 0) { dx+=x; x=0; }
	if (y < 0) { dy+=y; y=0; }
	if (x+dx > w) dx=w-x;
	if (y+dy > h) dy=h-y;

	// anything left to fill?

	if (dx<=0 || dy<=0) return;

	// compute coordinates

	dst = fa->dst.Address32(x, y+dy-1);

	// do the fill

	do {
		PixDim dxt = dx;
		Pixel32 *dst2 = dst;

		do {
			*dst2++ = c;
		} while(--dxt);

		dst = (Pixel32 *)((char *)dst + pitch);
	} while(--dy);

return;

}
